import time
from loguru import logger

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class Retry:
    """
    重试工具
    """

    def __init__(self, max_attempts=3, delay=1,
                 log_enable: bool = True,
                 ever_fail_level: Literal['TRACE', 'DEBUG', 'INFO', 'WARNING', 'ERROR'] = 'WARN',
                 fail_level: Literal['TRACE', 'DEBUG', 'INFO', 'WARNING', 'ERROR'] = 'ERROR'):
        """
        初始化重试工具
        Args:
            max_attempts (int): 最大重试次数，默认为 3 次。
            delay (float): 重试之间的延迟时间，单位为秒，默认为 1 秒。
            log_enable: 是否记录日志
            ever_fail_level: 每次失败记录的日志级别
            fail_level: 全部失败记录的日志级别
        """
        self._max_attempts = max_attempts
        self._delay = delay
        self._log_enable = log_enable
        self._ever_fail_level = ever_fail_level
        self._fail_level = fail_level

    def do(self, func, *args, **kwargs):
        """
        重试函数，重复执行指定的函数，直到达到最大重试次数或者函数执行成功为止。
        Args:
            func: 要执行的函数。
            args: 执行函数的参数
            kwargs: 执行函数的参数
        Returns:
            函数执行成功时返回函数的返回值，否则返回 None。
        """
        for attempt in range(1, self._max_attempts + 1):
            try:
                result = func(*args, **kwargs)
                return result  # 函数执行成功，直接返回结果
            except Exception as e:
                if attempt < self._max_attempts:
                    if self._log_enable:
                        logger.log(self._ever_fail_level,
                                   "函数[{}]第[{}]次执行失败, 失败原因: {}, 等待[{}]秒后将进行重试, 函数参数args: {}, kwargs: {}",
                                   func.__name__, attempt, e, self._delay, args, kwargs)
                    time.sleep(self._delay)
                else:
                    if self._log_enable:
                        logger.log(self._fail_level,
                                   '函数[{}]重试[{}]次仍然失败, 抛出异常: {}, 函数参数args: {}, kwargs: {}',
                                   func.__name__, self._max_attempts, e, args, kwargs)
                    raise e



import os
import smtplib
from email.header import Header
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.base import MIMEBase
from email.utils import formataddr
from email import encoders
from typing import Union, List, Tuple


class SmtpEmailNotice:
    """
    邮件通知工具
    """

    def __init__(self,
                 smtp_server: str,
                 smtp_port: int,
                 from_email: str,
                 from_password: str,
                 from_name: str = None,
                 ssl=True):
        """
        :param smtp_server: SMTP 服务器域名, 例如: smtp.feishu.cn
        :param smtp_port: SMTP 服务器端口, 例如: 465
        :param from_email: 发件人邮箱, 例如: xxx@xxx.com
        :param from_password: 发件人密码
        :param from_name: 发件人名称
        :param ssl: 是否使用SMTP_SSL连接服务器
        """
        self._smtp_server = smtp_server
        self._smtp_port = smtp_port
        self._from_email = from_email
        self._from_password = from_password
        self._from_name = from_name if from_name else from_email
        self._ssl = ssl

    def send_email(self,
                   title: str,
                   body: str,
                   to_emails: Union[List[str], Tuple[str], str],
                   to_cc_mails: Union[List[str], Tuple[str], str, None] = None,
                   attachments: Union[List[str], Tuple[str], str, None] = None):
        """
        发送邮件功能, 支持抄送、附件, 目前仅支持 html 渲染的 body 样式
        :param title: 邮件主题
        :param body: 邮件正文
        :param to_emails: 收件人, 支持多个收件人
        :param to_cc_mails: 抄送人, 支持多个抄送人
        :param attachments: 附件, 支持多个附件, 附件的绝对路径或相对路径
        :raise SMTPException: 发送邮件异常
        """
        message = MIMEMultipart()
        # 主题
        message['Subject'] = Header(title, 'utf-8')
        # 发件人
        message['From'] = formataddr(
            (self._from_name, self._from_email), 'utf-8')
        # 抄送人
        if isinstance(to_cc_mails, (tuple, list)):
            to_cc_mails = ','.join(to_cc_mails)
        message['Cc'] = to_cc_mails
        # 邮件正文
        message.attach(MIMEText(body, 'html'))
        # 附件
        if attachments:
            if isinstance(attachments, str):
                attachments = [attachments]
            for attachment in attachments:
                with open(attachment, 'rb') as f:
                    part = MIMEBase('application', 'octet-stream')
                    part.set_payload(f.read())
                encoders.encode_base64(part)
                filename = attachment.split(os.sep)[-1]
                part.add_header('Content-Disposition',
                                f'attachment; filename= {filename}')
                message.attach(part)

        # 连接服务器发送邮件
        if self._ssl:
            server = smtplib.SMTP_SSL(self._smtp_server, self._smtp_port)
        else:
            server = smtplib.SMTP(self._smtp_server, self._smtp_port)
        server.login(self._from_email, self._from_password)
        server.send_message(message, self._from_email, to_emails)



from typing import Union

from .base import DataFileLoader


class EvalDataFileLoader(DataFileLoader):

    def _convert_config_data_to_python(self,
                                       config_data: Union[str, bytes],
                                       **kwargs) -> Union[dict, tuple, list]:
        return eval(config_data)



from typing import Union

from .base import DataFileLoader


def _load_properties_data(config_data: str):
    def dfs_properties_key(p, keys, value):
        if len(keys) == 1:
            p[keys[0]] = value
            return
        if keys[0] not in p:
            p[keys[0]] = {}
        dfs_properties_key(p[keys[0]], keys[1:], value)

    properties = {}
    for line in config_data.split('\n'):
        if not line.strip() or line.startswith('#') or '=' not in line:
            continue
        key, value = line.split('=', 1)
        key, value = key.strip(), value.strip()
        keys = key.split(".")
        if len(keys) >= 2:
            dfs_properties_key(properties, keys, value)
        else:
            properties[key] = value

    return properties


class PropertiesDataFileLoader(DataFileLoader):

    def _convert_config_data_to_python(self,
                                       config_data: Union[str, bytes],
                                       encoding: str = 'utf-8',
                                       **kwargs) -> Union[dict, tuple, list]:
        if isinstance(config_data, bytes):
            config_data = config_data.decode(encoding=encoding)
        return _load_properties_data(config_data)



from loguru import logger as log
from dragons96_tools.logger import setup as stp

stp('test_logger.log')


@log.catch
def error_fun():
    return 1 / 0


def test_log_catch():
    f = error_fun()
    print(f)




from sqlalchemy import create_engine, Engine
from sqlalchemy.orm import sessionmaker, DeclarativeBase, Session
from sqlalchemy.ext.asyncio import create_async_engine, async_sessionmaker, AsyncEngine, AsyncSession
from contextlib import contextmanager, asynccontextmanager
from typing import Any, Type, TypeVar
from pydantic import BaseModel

_T = TypeVar('_T', bound='BaseModel')
_BASE = TypeVar('_BASE', bound='Base')


class Base(DeclarativeBase):
    """ orm model 基础类 """

    def to_dict(self) -> dict:
        """
        将当前model转为dict对象
        """
        return {col.name: getattr(self, col.name) for col in self.__table__.columns}

    def to_model(self, modelclass: Type[BaseModel]) -> _T:
        """
        将当前model转为pydantic的model对象
        :param modelclass: 具体的模型类型
        """
        return modelclass(**(self.to_dict()))

    @classmethod
    def from_model(cls,
                   model: BaseModel,
                   ignore_absent_fields: bool = True,
                   include: Any = None,
                   exclude: Any = None,
                   by_alias: bool = False,
                   exclude_unset: bool = False,
                   exclude_defaults: bool = False,
                   exclude_none: bool = False,
                   round_trip: bool = False,
                   ) -> _BASE:
        """
        将pydantic的model转为当前类型model
        :param model: pydantic的模型值
        :param ignore_absent_fields: 忽略当前model中不存在的pydantic model的属性
        :param include: A list of fields to include in the output.
        :param exclude: A list of fields to exclude from the output.
        :param by_alias: Whether to use the field's alias in the dictionary key if defined.
        :param exclude_unset: Whether to exclude fields that have not been explicitly set.
        :param exclude_defaults: Whether to exclude fields that are set to their default value.
        :param exclude_none: Whether to exclude fields that have a value of `None`.
        :param round_trip: If True, dumped values should be valid as input for non-idempotent types such as Json[T].
        """
        model_dict = model.model_dump(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip
        )
        if ignore_absent_fields:
            for key in list(model_dict.keys()):
                if not hasattr(cls, key):
                    del model_dict[key]
        return cls(**model_dict)


class SqlAlchemyClient:
    """
    SqlAlchemy工具客户端
    """
    def __init__(self, url: str, echo: bool = False, **kwargs: Any):
        """
        创建sqlalchemy工具客户端
        Args:
            url: 连接串
                示例:
                    sqlite:///dbname
                    mysql+pymysql://username:password@host:port/dbname
            echo: The echo=True parameter indicates that SQL emitted by connections will be logged to standard out.
        """
        self._engine = create_engine(url, echo=echo, **kwargs)
        self._session_maker = sessionmaker(bind=self._engine, expire_on_commit=False)

    @contextmanager
    def session(self) -> Session:
        """
        获取session上下文对象
        使用示例:
            helper = SqlAlchemyHelper('xxxx')
            with helper.session() as session:
                session.execute()
                session.get()
        """
        session = self._session_maker()
        try:
            yield session
            session.commit()
        except Exception as e:
            session.rollback()
            raise e
        finally:
            if session:
                session.close()

    def engine(self) -> Engine:
        """ 获取引擎 """
        return self._engine


class AsyncSqlAlchemyClient:
    """
    SqlAlchemy异步工具客户端
    """
    def __init__(self, url: str, echo: bool = False, **kwargs: Any):
        """
        创建sqlalchemy异步工具客户端
        :param url: 连接串
            示例:
                sqlite:///dbname
                mysql+pymysql://username:password@host:port/dbname
        :param echo: The echo=True parameter indicates that SQL emitted by connections will be logged to standard out.
        """
        self._engine = create_async_engine(url, echo=echo, **kwargs)
        self._session_maker = async_sessionmaker(bind=self._engine)

    @asynccontextmanager
    async def session(self) -> AsyncSession:
        """
        获取session上下文对象
        使用示例:
            helper = SqlAlchemyHelper('xxxx')
            with helper.session() as session:
                session.execute()
                session.get()
        """
        session = self._session_maker()
        try:
            yield session
            await session.commit()
        except Exception as e:
            await session.rollback()
            raise e
        finally:
            if session:
                await session.close()

    def engine(self) -> AsyncEngine:
        """ 获取引擎 """
        return self._engine



from typing import Optional, Union, Any
from .base import BaseBuilder


class MarkdownBuilder(BaseBuilder):
    """
    markdown 建造器
    """

    def __init__(self):
        self._md = ''
        self._tag_end = '\n\n'

    def h1(self, content: Union[str, 'MarkdownBuilder']):
        return self._before_sign('#', content)

    def h2(self, content: Union[str, 'MarkdownBuilder']):
        return self._before_sign('##', content)

    def h3(self, content: Union[str, 'MarkdownBuilder']):
        return self._before_sign('###', content)

    def h4(self, content: Union[str, 'MarkdownBuilder']):
        return self._before_sign('####', content)

    def h5(self, content: Union[str, 'MarkdownBuilder']):
        return self._before_sign('#####', content)

    def h6(self, content: Union[str, 'MarkdownBuilder']):
        return self._before_sign('######', content)

    def em(self, content: Union[str, 'MarkdownBuilder']):
        return self._wrapper_sign('*', content)

    def i(self, content: Union[str, 'MarkdownBuilder']):
        return self._wrapper_sign('_', content)

    def strong(self, content: Union[str, 'MarkdownBuilder']):
        return self._wrapper_sign('**', content)

    def b(self, content: Union[str, 'MarkdownBuilder']):
        return self._wrapper_sign('__', content)

    def del_(self, content: Union[str, 'MarkdownBuilder']):
        return self._wrapper_sign('~~', content)

    def a(self, content: Union[str, 'MarkdownBuilder'], href: str = ''):
        return self._text(f'[{content}]({href})')

    def img(self, src, alt=''):
        return self._text(f'![{alt}]({src})')

    def hr(self):
        return self._text('---')

    def blockquote(self, content: Union[str, 'MarkdownBuilder']):
        return self._before_sign('>', content)

    def code(self, content: Union[str, 'MarkdownBuilder']):
        return self._wrapper_sign('`', content)

    def pre(self, content: Union[str, 'MarkdownBuilder'], code_type=''):
        return self._wrapper_sign('```', f'{code_type}\n{content}\n')

    def p(self, content: Union[str, 'MarkdownBuilder']):
        return self.no_tag(content)

    def no_tag(self, content: Union[str, 'MarkdownBuilder']):
        return self._text(content)

    def _text(self, text: Union[str, 'MarkdownBuilder']):
        self._md += f'{text}{self._tag_end}'
        return self

    def _before_sign(self, sign: str, content: Union[str, 'MarkdownBuilder'] = ''):
        self._md += f'{sign} {content}{self._tag_end}'
        return self

    def _wrapper_sign(self, sign: str, content: Union[str, 'MarkdownBuilder'] = ''):
        self._md += f'{sign}{content}{sign}{self._tag_end}'
        return self

    def build(self):
        return self._md[:-2] if self._md else self._md



from dragons96_tools.models import BaseModel
from typing import Optional, Type, List, Tuple, Any, Union
import pymysql
from pymysql import connect as c, Connection
from pymysql.cursors import DictCursor, Cursor
from contextlib import contextmanager


def connect(host=None,
            port=3306,
            user=None,
            password='',
            database=None,
            cursorclass=DictCursor,
            autocommit=True,
            pool=False,
            pool_size=1,
            **kwargs):
    """
    业务Pymysql连接通用参数封装
    :return pymysql.Connection or dbutils.pooled_db.PooledDB
    """
    if pool:
        from dbutils.pooled_db import PooledDB
        return PooledDB(creator=connect,
                        mincached=pool_size,
                        host=host,
                        port=port,
                        user=user,
                        password=password,
                        database=database,
                        cursorclass=cursorclass,
                        autocommit=autocommit,
                        **kwargs)
    return c(host=host,
             port=port,
             user=user,
             password=password,
             database=database,
             cursorclass=cursorclass,
             autocommit=autocommit,
             **kwargs)


class _PyMysqlConfig(BaseModel):
    host: str
    port: int
    user: str
    password: Optional[str] = ''
    db: Optional[str] = None
    autocommit: Optional[bool] = False


class _PyMysqlTransactionCursor:
    """
    pymysql 事务 cursor 拓展
    仅执行 SQL 不提交不回滚, 由 PyMysqlHelper 处理提交回滚连接
    """

    def __init__(self, conn: Connection):
        self._conn = conn
        self._cursor = conn.cursor()

    def execute(self,
                sql: str,
                args: Union[List[Any], Tuple[Any], None] = None,
                modelclass: Optional[Type[BaseModel]] = None
                ):
        """
        以事务方式执行 sql, 保障同一个 对象 执行多次 execute 方法在同一事务中
        :param sql: 执行的sql
        :param args: 执行的sql中的参数
        :param modelclass: pydantic model 类型, 传入该值会将结果转为 pydantic model
        """
        self._cursor.execute(sql, args)
        ans = self._cursor.fetchall()
        if ans and modelclass:
            ans = list(map(lambda e: modelclass(**e), ans))
        return ans

    def close(self):
        if self._cursor:
            self._cursor.close()


class PyMysqlClient:
    """
    pymysql 帮助类
    """
    def __init__(self,
                 host='localhost',
                 port=3306,
                 user='root',
                 password='',
                 database=None,
                 **kwargs):
        self._config = _PyMysqlConfig(
            host=host,
            port=port,
            user=user,
            password=password,
            db=database
        )
        self._kw = kwargs

    def _new_connection(self):
        return c(host=self._config.host,
                 port=self._config.port,
                 user=self._config.user,
                 password=self._config.password,
                 db=self._config.db,
                 autocommit=self._config.autocommit,
                 **self._kw)

    def execute(self,
                sql: str,
                args: Union[List[Any], Tuple[Any], None] = None,
                modelclass: Optional[Type[BaseModel]] = None) -> Union[
        List[dict], List[BaseModel], Tuple[dict], Tuple[BaseModel]]:
        """
        执行 非事务模式 sql
        :param sql: 执行的sql
        :param args: 执行的sql中的参数
        :param modelclass: pydantic model 类型, 传入该值会将结果转为 pydantic model
        """
        conn = self._new_connection()
        cursor = conn.cursor(cursor=DictCursor)
        try:
            cursor.execute(sql, args)
            conn.commit()
            ans = cursor.fetchall()
            if ans and modelclass:
                ans = list(map(lambda e: modelclass(**e), ans))
            return ans
        except Exception as e:
            conn.rollback()
            raise e
        finally:
            if cursor:
                cursor.close()
            if conn:
                conn.close()

    @contextmanager
    def transaction_client(self) -> _PyMysqlTransactionCursor:
        """
        获取事务游标对象
        使用方式
            helper = PyMysqlHelper(...)
            with helper.cursor() as cursor:
                users = cursor.execute('select * from user')
                cursor.execute('update user set user_name= "sss" where user_id=1')

        """
        conn = self._new_connection()
        transaction_cursor = _PyMysqlTransactionCursor(conn)
        try:
            yield transaction_cursor
            conn.commit()
        except Exception as e:
            conn.rollback()
            raise e
        finally:
            transaction_cursor.close()
            if conn:
                conn.close()



from loguru import logger


def setup(file_name,
          rotation="1 days",
          retention="30 days",
          serialize=True,
          backtrace=True,
          diagnose=False,
          filter=lambda record: True,
          level='INFO'):
    """
    初始化安装loguru日志配置
    使用方式:
        from loguru import logger

        logger.info('info message')
        logger.error('error message')
    :param file_name: 文件名称
    :param rotation: 日志滚动周期, 允许值示例如下:
        logger.add("file_1.log", rotation="500 MB")    # Automatically rotate too big file
        logger.add("file_2.log", rotation="12:00")     # New file is created each day at noon
        logger.add("file_3.log", rotation="1 week")    # Once the file is too old, it's rotated
        logger.add("file_X.log", retention="10 days")  # Cleanup after some time
        logger.add("file_Y.log", compression="zip")    # Save some loved space
    :param retention: 日志文件保留周期, 默认30天
    :param serialize: 是否将写入文件的数据序列化为json, 默认True. 推荐: True
    :param backtrace: 是否默认在异常时自动打印堆栈信息, 默认True. 推荐: True
    :param diagnose: 是否开启代码异常诊断, 可层层递进追踪异常链路, 生产环境建议关闭, 开发环境建议开启, 默认False
    :param filter: 日志文件的过滤条件, 符合条件的日志记录将写入该文件
    :param level: 日志级别, 只有超过该级别的日志会记入文件, 支持以下日志级别
        TRACE: 最低级别，用于追踪程序的详细执行流程，通常用于诊断问题
        DEBUG: 用于调试目的的详细信息，例如变量值或者函数调用
        INFO: 提供程序执行中的一般信息，用于表示程序正常运行
        SUCCESS: 表示成功的操作
        WARNING: 表示潜在的问题或者不符合预期的情况，但程序仍能继续执行
        ERROR: 表示错误，但程序仍然可以继续执行
        CRITICAL: 表示严重错误，可能导致程序无法继续执行
    """
    # 添加文件配置
    logger.add(file_name,
               format="{message}",
               backtrace=backtrace,
               diagnose=diagnose,
               serialize=serialize,
               rotation=rotation,
               retention=retention,
               filter=filter,
               level=level)


def get_loguru_adapter_logging_formatter():
    """
    获取适配loguru serialize格式的logging.Formatter类型
    """
    import logging
    import time
    import traceback
    import datetime
    import json

    class LoguruSerializeAdapterFormatter(logging.Formatter):
        def format(self, record):
            elapsed_seconds = time.time() - record.created
            elapsed_repr = self._format_elapsed(elapsed_seconds)
            exception_msg = traceback.format_exc() if record.exc_info else ''
            log_record = {
                "text": record.getMessage() + f"\n{exception_msg}",
                "record": {
                    "elapsed": {
                        "repr": elapsed_repr,
                        "seconds": elapsed_seconds
                    },
                    "exception": self._format_exception(record.exc_info),
                    "extra": {},
                    "file": {
                        "name": record.filename,
                        "path": record.pathname
                    },
                    "function": record.funcName,
                    "level": {
                        "icon": self._get_level_icon(record.levelname),
                        "name": record.levelname,
                        "no": record.levelno
                    },
                    "line": record.lineno,
                    "message": record.getMessage(),
                    "module": record.module,
                    "name": record.name,
                    "process": {
                        "id": record.process,
                        "name": record.processName
                    },
                    "thread": {
                        "id": record.thread,
                        "name": record.threadName
                    },
                    "time": {
                        "repr": datetime.datetime.fromtimestamp(record.created).strftime('%Y-%m-%d %H:%M:%S.%f%z'),
                        "timestamp": record.created
                    }
                }
            }
            return json.dumps(log_record)

        @staticmethod
        def _get_level_icon(level_name: str):
            if level_name == 'DEBUG':
                return '🐞'
            elif level_name == 'INFO':
                return 'ℹ️'
            elif level_name == 'WARNING':
                return '⚠️'
            elif level_name == 'ERROR':
                return '❌'
            elif level_name == 'CRITICAL':
                return '☠️'
            return ''

        @staticmethod
        def _format_elapsed(elapsed_seconds):
            hours, remainder = divmod(elapsed_seconds, 3600)
            minutes, seconds = divmod(remainder, 60)
            return "{:0>2}:{:0>2}:{:0>2}.{:06}".format(int(hours), int(minutes), int(seconds),
                                                       int((elapsed_seconds - int(elapsed_seconds)) * 1000000))

        @staticmethod
        def _format_exception(exc_info):
            if exc_info:
                exc_type, exc_value, exc_traceback = exc_info
                return {
                    "type": exc_type.__name__,
                    "message": str(exc_value),
                    "traceback": True
                }
            else:
                return None

    return LoguruSerializeAdapterFormatter



from fastapi import FastAPI
from fastapi.responses import PlainTextResponse
from fastapi.exceptions import RequestValidationError, HTTPException
from dragons96_tools.models import R
from loguru import logger


def wrapper_exception_handler(app: FastAPI) -> FastAPI:
    """
    给FastAPI对象增加通用异常处理
    """

    @app.exception_handler(RequestValidationError)
    async def request_validation_exception_handler(request, exc: RequestValidationError):
        logger.error('发生请求参数校验异常, {}', exc)
        return PlainTextResponse(R.fail(msg='参数校验不通过', code=400).model_dump_json())

    @app.exception_handler(AssertionError)
    async def assert_exception_handler(request, exc):
        logger.error('发生断言异常, {}', exc)
        return PlainTextResponse(R.fail(msg=str(exc)).model_dump_json())

    @app.exception_handler(HTTPException)
    async def http_exception_handler(request, exc: HTTPException):
        logger.error('发生http异常, {}', exc)
        return PlainTextResponse(str(exc.detail), status_code=exc.status_code)

    @app.exception_handler(404)
    async def exception_404_handler(request, exc):
        logger.error('[404]请求不存在的路径, {}', request.url.path)
        return PlainTextResponse(R.fail(msg='资源不存在', code=404).model_dump_json())

    @app.exception_handler(Exception)
    async def exception_handler(request, exc):
        logger.error('发生未知异常, {}', exc)
        return PlainTextResponse(R.fail(msg='内部服务器错误').model_dump_json())

    return app



import json
import time
from typing import Optional, Any, Union, Dict, List
import pydantic
import requests
from ..models import BaseModel
from ..cryptography import sha256_hmac, encode_base64
import re


class _FeishuRobotBaseMsg(BaseModel):

    def get_msg_type(self):
        raise NotImplementedError()

    def to_json(self):
        return self.model_dump(exclude_none=True)


class FeishuRobotTextMsg(_FeishuRobotBaseMsg):
    """
    飞书机器人文本消息封装
    """
    # 发送的文本内容
    text: str

    def get_msg_type(self):
        return 'text'


class FeishuRobotPostMsgTag(BaseModel):
    """
    飞书机器人富文本消息标签封装
    """
    # 标签类型, 支持 text(文本), a(链接), at(艾特)
    tag: str
    # 文本消息
    text: Optional[str] = None
    # 表示是否 unescape 解码。默认值为 false，未用到 unescape 时可以不填
    un_escape: Optional[bool] = None
    # 链接地址
    href: Optional[str] = None
    # at目标的ID
    user_id: Optional[str] = None
    # at目标的用户名称
    user_name: Optional[str] = None
    """
    图片的唯一标识。可通过 上传图片 接口获取 image_key
    https://open.feishu.cn/document/uAjLw4CM/ukTMukTMukTM/reference/im-v1/image/create
    """
    image_key: Optional[str] = None


_FeishuRobotPostMsgContent = list


class FeishuRobotPostMsgContentBuilder:
    """
    飞书富文本消息内容建造器
    """

    def __init__(self):
        self._data = [[]]
        self._index = 0

    def add_text(self, text: str, un_escape: Optional[bool] = None):
        """
        添加一个text文本标签
        :param text: 文本内容
        :param un_escape: 表示是否 unescape 解码。默认值为 false，未用到 unescape 时可以不填
        """
        self._data[self._index].append(
            FeishuRobotPostMsgTag(tag='text', text=text, un_escape=un_escape).model_dump(exclude_none=True))
        return self

    def add_a(self, text: str, href: str):
        """
        添加一个a超链接标签
        :param text: 超链接的文本内容
        :param href: 默认的链接地址，你需要确保链接地址的合法性，否则消息会发送失败
        """
        self._data[self._index].append(
            FeishuRobotPostMsgTag(tag='a', text=text, href=href).model_dump(exclude_none=True))
        return self

    def add_at(self, user_id: str, user_name: Optional[str] = None):
        """
        添加一个at标签
        :param user_id:
            用户的 Open ID 或 User ID。
            - @ 单个用户时，user_id字段必须是有效值（仅支持 @ 自定义机器人所在群的群成员）。
            - @ 所有人时，填 all。
        :param user_name:
            用户名称
        """
        self._data[self._index].append(
            FeishuRobotPostMsgTag(tag='at', user_id=user_id, user_name=user_name).model_dump(exclude_none=True))
        return self

    def add_img(self, image_key):
        """
        添加一个img标签
        :param image_key:
            图片的唯一标识。可通过 上传图片 接口获取 image_key
            https://open.feishu.cn/document/uAjLw4CM/ukTMukTMukTM/reference/im-v1/image/create
        """
        self._data[self._index].append(
            FeishuRobotPostMsgTag(tag='img', image_key=image_key).model_dump(exclude_none=True))
        return self

    def add_newline(self):
        """
        换行
        """
        self._data.append([])
        self._index += 1
        return self

    def build(self) -> _FeishuRobotPostMsgContent:
        """
        生成 FeishuRobotPostMsg.content 字段对象
        """
        return self._data


class FeishuRobotPostMsg(_FeishuRobotBaseMsg):
    """
    飞书富文本消息封装
    usage:
        FeishuRobotPostMsg(title='xxx',
                           content=FeishuRobotPostMsg.content_builder().add_text('xxxx').build()
        )
    """
    # 标题
    title: Optional[str] = None
    # 内容
    content: _FeishuRobotPostMsgContent

    def get_msg_type(self):
        return 'post'

    def to_json(self, lang='zh_cn'):
        return {'post': {lang: self.model_dump()}}

    @staticmethod
    def content_builder():
        return FeishuRobotPostMsgContentBuilder()


class FeishuRobotCardMsg(_FeishuRobotBaseMsg):
    """
    飞书机器人卡片消息
    建议使用飞书消息卡片搭建工具快速创建模板, 直接构建对象传参
    https://open.feishu.cn/tool/cardbuilder?templateId=ctp_AAYHcTkBGx0k
    """
    # 卡片消息体内容
    elements: list
    # 卡片消息配置
    config: Optional[dict] = None
    # 卡片消息标题
    header: Optional[dict] = None

    def get_msg_type(self):
        return 'interactive'


class FeishuRobotCardMsgTemplate(BaseModel):
    """
    飞书机器人卡片消息模板, 模板规则, 使用 {{}} 包裹的文本视为变量, 例如 {{name}} 表示变量name
    """
    # 卡片消息文本
    elements: list
    # 卡片消息配置
    config: Optional[dict] = None
    # 卡片消息标题
    header: Optional[dict] = None

    def gen_msg(self, elements_map: Union[Dict[str, Any], pydantic.BaseModel, None] = None,
                header_map: Union[Dict[str, Any], pydantic.BaseModel, None] = None) -> FeishuRobotCardMsg:
        """
        通过模板生成消息配置
        :param elements_map: 卡片消息模板名称与值的映射
        :param header_map: 卡片消息标题模板名称与值的映射
        """
        elements = json.dumps(self.elements, ensure_ascii=False)
        if elements_map:
            if isinstance(elements_map, pydantic.BaseModel):
                elements_map = elements_map.model_dump()
            for name, value in elements_map.items():
                elements = re.sub(
                    '{{\\s*?' + name + '\\s*?}}', str(value), elements)
        header = json.dumps(
            self.header, ensure_ascii=False) if self.header else None
        if header and header_map:
            if isinstance(header_map, pydantic.BaseModel):
                header_map = header_map.model_dump()
            for name, value in header_map.items():
                header = re.sub('{{\\s*?' + name + '\\s*?}}',
                                str(value), header)
        if elements:
            elements = json.loads(elements)
        if header:
            header = json.loads(header)
        return FeishuRobotCardMsg(elements=elements, config=self.config, header=header)


class _FeishuRobotFullMsg(BaseModel):
    # 消息类型
    msg_type: str
    # 普通文本、富文本消息内容
    content: Optional[Any] = None
    # 卡片消息内容 (与content二选一)
    card: Optional[Any] = None
    # 传递secret秘钥时必传时间戳校验
    timestamp: Optional[str] = None
    # 传递secret秘钥时必传校验sign
    sign: Optional[str] = None

    def set_msg(self, msg: _FeishuRobotBaseMsg):
        if isinstance(msg, (FeishuRobotTextMsg, FeishuRobotPostMsg)):
            self.content = msg.to_json()
        if isinstance(msg, FeishuRobotCardMsg):
            self.card = msg.to_json()


class FeishuRobotNotice:
    """
    飞书机器人通知工具
    """

    def __init__(self, webhook: str, secret: Optional[str] = None):
        """
        :param webhook: 飞书webhook地址
        :param secret: 签名校验秘钥, 非必填
        """
        self._webhook = webhook
        self._secret = secret

    def _gen_sign(self, timestamp: Union[int, str]):
        """
        生成飞书的sign
        :param timestamp: 秒级时间戳
        """
        # 先拼接串
        sec = f'{timestamp}\n{self._secret}'
        # base64 + sha256 hmac 编码
        return encode_base64(sha256_hmac(sec, encoding='utf-8')).decode('utf-8')

    def send_text_msg(self, text: str, timeout: Union[float, tuple] = None):
        """
        发送飞书机器人纯文本消息
        :param text: 文本消息
        :param timeout: 可选, 发送消息超时时间, 支持float或者tuple2 (connection timeout, read timeout)
        """
        msg = FeishuRobotTextMsg(text=text)
        return self.send_msg(msg, timeout)

    def send_post_msg(self,
                      title: Optional[str] = None,
                      content_builder: FeishuRobotPostMsgContentBuilder = None,
                      timeout: Union[float, tuple] = None):
        """
        发送飞书机器人富文本消息
        :param title: 富文本标题
        :param content_builder: 富文本内容建造器, 通过 FeishuRobotPostMsg.content_builder() 可获取Builder实例
        :param timeout: 可选, 发送消息超时时间, 支持float或者tuple2 (connection timeout, read timeout)
        """
        content = []
        if content_builder:
            content = content_builder.build()
        msg = FeishuRobotPostMsg(title=title, content=content)
        return self.send_msg(msg, timeout)

    def send_msg(self, msg: _FeishuRobotBaseMsg, timeout: Union[float, tuple] = None):
        """
        发送飞书机器人消息
        :param msg: 飞书机器人消息
        :param timeout: 可选, 发送消息超时时间, 支持float或者tuple2 (connection timeout, read timeout)
        """
        # 初始化完整消息对象
        full_msg = _FeishuRobotFullMsg(msg_type=msg.get_msg_type())
        # 设置业务消息
        full_msg.set_msg(msg)
        # 有秘钥则补充sign校验相关参数
        if self._secret:
            full_msg.timestamp = str(int(time.time()))
            full_msg.sign = self._gen_sign(full_msg.timestamp)
        response = requests.post(self._webhook,
                                 json=full_msg.model_dump(exclude_none=True),
                                 timeout=timeout)
        if response.status_code != 200:
            raise Exception(
                '发送飞书机器人消息响应状态码异常, 状态码: {}, 响应信息: {}', response.text)
        return response.json()



from typing import Union
from enum import Enum
from .base import BaseBuilder


class ATarget(Enum):
    SELF = '_self'
    BLANK = '_blank'
    PARENT = '_parent'
    TOP = '_top'


class HtmlBuilder(BaseBuilder):
    """
    HTML 标签建造器
    """

    def __init__(self, auto_add_html_tag=False):
        self.html = ''
        self._tag_end = '\n'
        self._auto_add_html_tag = auto_add_html_tag

    def h1(self, content_: Union[str, 'HtmlBuilder'], **attrs):
        return self._common_tag('h1', content_, **attrs)

    def h2(self, content_: Union[str, 'HtmlBuilder'], **attrs):
        return self._common_tag('h2', content_, **attrs)

    def h3(self, content_: Union[str, 'HtmlBuilder'], **attrs):
        return self._common_tag('h3', content_, **attrs)

    def h4(self, content_: Union[str, 'HtmlBuilder'], **attrs):
        return self._common_tag('h4', content_, **attrs)

    def h5(self, content_: Union[str, 'HtmlBuilder'], **attrs):
        return self._common_tag('h5', content_, **attrs)

    def h6(self, content_: Union[str, 'HtmlBuilder'], **attrs):
        return self._common_tag('h6', content_, **attrs)

    def br(self):
        return self._common_tag('br')

    def b(self, content_: Union[str, 'HtmlBuilder'], **attrs):
        return self._common_tag('b', content_, **attrs)

    def em(self, content_: Union[str, 'HtmlBuilder'], **attrs):
        return self._common_tag('em', content_, **attrs)

    def i(self, content_: Union[str, 'HtmlBuilder'], **attrs):
        return self._common_tag('i', content_, **attrs)

    def blockquote(self, content_: Union[str, 'HtmlBuilder'], **attrs):
        return self._common_tag('blockquote', content_, **attrs)

    def code(self, content_: Union[str, 'HtmlBuilder'], **attrs):
        return self._common_tag('code', content_, **attrs)

    def pre(self, content_: Union[str, 'HtmlBuilder'], **attrs):
        return self._common_tag('pre', content_, **attrs)

    def p(self, content_: Union[str, 'HtmlBuilder'] = None, **attrs):
        return self._common_tag('p', content_, **attrs)

    def a(self, content_: Union[str, 'HtmlBuilder'] = None, href: str = '', target: Union[ATarget, str] = ATarget.SELF, **attrs):
        if isinstance(target, ATarget):
            target = target.value
        return self._common_tag('a', content_, href=href, target=target, **attrs)

    def img(self, src, **attrs):
        return self._common_tag('img', src=src, **attrs)

    def strong(self, content_: Union[str, 'HtmlBuilder'], **attrs):
        return self._common_tag('strong', content_, **attrs)

    def div(self, content_: Union[str, 'HtmlBuilder'] = None, **attrs):
        return self._common_tag('div', content_, **attrs)

    def no_tag(self, content_: Union[str, 'HtmlBuilder']):
        """无标签"""
        return self._common_tag(content_=content_)

    def custom_tag(self, tag_: str, content_: Union[str, 'HtmlBuilder'] = None, **attrs):
        """自定义标签"""
        return self._common_tag(tag_, content_, **attrs)

    def _common_tag(self, tag_: str = None, content_: Union[str, 'HtmlBuilder'] = None, **attrs):
        attrs_str = self._convert_attrs_to_str(**attrs)
        if content_ and tag_:
            self.html += f'<{tag_}{attrs_str}>{content_}</{tag_}>{self._tag_end}'
            return self
        if content_:
            self.html += f'{content_}{self._tag_end}'
            return self
        if tag_:
            self.html += f'<{tag_}{attrs_str} />{self._tag_end}'
            return self
        return self

    @staticmethod
    def _convert_attrs_to_str(**attrs):
        attrs_str = ''
        if attrs:
            for key, value in attrs.items():
                attrs_str += f' {key}="{value}"'
        return attrs_str

    def build(self):
        html = self.html
        if self._auto_add_html_tag:
            html = f'<html>\n{html}\n</html>'
        return html



from dragons96_tools.retry import Retry
from loguru import logger

cnt = 0


def fail3(a: int, b: int):
    logger.info('请求参数a: {}, b: {}', a, b)
    global cnt
    if cnt < 3:
        cnt += 1
        raise Exception('异常')
    logger.info('cnt: {}', cnt)
    return cnt


def test_retry():
    _retry = Retry(max_attempts=10,  # 最大重试次数
                   delay=1,  # 每次重试的等待间隔, 单位秒
                   log_enable=True,  # 是否打印日志
                   ever_fail_level='WARNING',  # 每次任务失败记录的日志级别(未达到重试上限),
                   fail_level='ERROR',  # 任务重试达到失败上限的日志级别
                   )
    logger.info(_retry.do(fail3, 1, b=5))



import hashlib
import hmac as hmac
from typing import Union
from .base import _convert_data


def _common_hmac(key: Union[str, bytes],
                 data: Union[str, bytes, None] = None,
                 digestmod: Union[str, callable] = '',
                 encoding='utf-8') -> bytes:
    key = _convert_data(key, encoding=encoding)
    if data:
        data = _convert_data(data, encoding=encoding)
    return hmac.new(key, data, digestmod).digest()


def md5_hmac(key: Union[str, bytes],
             data: Union[str, bytes, None] = None,
             encoding='utf-8') -> bytes:
    """
    MD5 HMAC 加密
    """
    return _common_hmac(key, data, hashlib.md5,
                        encoding=encoding)


def sha1_hmac(key: Union[str, bytes],
              data: Union[str, bytes, None] = None,
              encoding='utf-8') -> bytes:
    """
    SHA1 HMAC 加密
    """
    return _common_hmac(key, data, hashlib.sha1,
                        encoding=encoding)


def sha224_hmac(key: Union[str, bytes],
                data: Union[str, bytes, None] = None,
                encoding='utf-8') -> bytes:
    """
    SHA224 HMAC
    """
    return _common_hmac(key, data, hashlib.sha224,
                        encoding=encoding)


def sha256_hmac(key: Union[str, bytes],
                data: Union[str, bytes, None] = None,
                encoding='utf-8') -> bytes:
    """
    SHA256 HMAC 加密
    """
    return _common_hmac(key, data, hashlib.sha256,
                        encoding=encoding)


def sha384_hmac(key: Union[str, bytes],
                data: Union[str, bytes, None] = None,
                encoding='utf-8') -> bytes:
    """
    SHA384 HMAC 加密
    """
    return _common_hmac(key, data, hashlib.sha384,
                        encoding=encoding)


def sha512_hmac(key: Union[str, bytes],
                data: Union[str, bytes, None] = None,
                encoding='utf-8') -> bytes:
    """
    SHA512 HMAC 加密
    """
    return _common_hmac(key, data, hashlib.sha512,
                        encoding=encoding)



from typing import Union, Optional, Any
import base64
from .base import _convert_data
import sys

_encode_handler_map = {
    '16': base64.b16encode,
    '32': base64.b32encode,
    '64': base64.b64encode,
    '85': base64.b85encode,
}

_decode_handler_map = {
    '16': base64.b16decode,
    '32': base64.b32decode,
    '64': base64.b64decode,
    '85': base64.b85decode,
}


def _common_encode_base(data: Union[str, bytes],
                        encoding='utf-8',
                        digit: str = '64') -> bytes:
    data = _convert_data(data, encoding)
    return _encode_handler_map[digit](data)


def _common_decode_base(data: Union[str, bytes],
                        encoding='utf-8',
                        digit: str = '64',
                        *,
                        casefold=False,
                        map01=None,
                        altchars=None,
                        validate=False) -> bytes:
    data = _convert_data(data, encoding)
    kwargs = {}
    if casefold:
        kwargs['casefold'] = True
    if map01:
        kwargs['map01'] = map01
    if altchars:
        kwargs['altchars'] = altchars
    if validate:
        kwargs['validate'] = validate
    return _decode_handler_map[digit](data, **kwargs)


def encode_base16(data: Union[str, bytes],
                  encoding='utf-8') -> bytes:
    """
    BASE16 编码
    """
    return _common_encode_base(data, encoding, '16')


def encode_base32(data: Union[str, bytes],
                  encoding='utf-8') -> bytes:
    """
    BASE32 编码
    """
    return _common_encode_base(data, encoding, '32')


def encode_base64(data: Union[str, bytes],
                  encoding='utf-8') -> bytes:
    """
    BASE64 编码
    """
    return _common_encode_base(data, encoding, '64')


def encode_base85(data: Union[str, bytes],
                  encoding='utf-8') -> bytes:
    """
    BASE85 编码
    """
    return _common_encode_base(data, encoding, '85')


def decode_base16(data: Union[str, bytes],
                  encoding='utf-8',
                  casefold=False) -> bytes:
    """
    BASE16 解码
    """
    return _common_decode_base(data, encoding, '16',
                               casefold=casefold)


def decode_base32(data: Union[str, bytes],
                  encoding='utf-8',
                  casefold=False,
                  map01: Optional[bytes] = None) -> bytes:
    """
    BASE32 解码
    """
    return _common_decode_base(data, encoding, '32',
                               casefold=casefold,
                               map01=map01)


def decode_base64(data: Union[str, bytes],
                  encoding='utf-8',
                  altchars: Any = None,
                  validate: bool = False) -> bytes:
    """
    BASE64 解码
    """
    return _common_decode_base(data, encoding, '64',
                               altchars=altchars,
                               validate=validate)


def decode_base85(data: Union[str, bytes],
                  encoding='utf-8') -> bytes:
    """
    BASE85 解码
    """
    return _common_decode_base(data, encoding, '85')


if sys.version_info >= (3, 10):
    _encode_handler_map['hex32'] = base64.b32hexencode
    _decode_handler_map['hex32'] = base64.b32hexdecode

    def encode_base32hex(data: Union[str, bytes],
                         encoding='utf-8') -> bytes:
        """
        BASE32 HEX 编码
        """
        return _common_encode_base(data, encoding, 'hex32')

    def decode_base32hex(data: Union[str, bytes],
                         encoding='utf-8',
                         casefold=False) -> bytes:
        """
        BASE32 HEX 解码
        """
        return _common_decode_base(data, encoding, 'hex32',
                                   casefold=casefold)
